@@ -36,9 +36,10 @@
 
 -include_lib("osmo_gsup/include/gsup_protocol.hrl").
 -include_lib("gtplib/include/gtp_packet.hrl").
+-include_lib("gtp_utils.hrl").
 -include("conv.hrl").
 
--export([start_link/1, stop/1]).
+-export([start_monitor/1, stop/1]).
 -export([init/1,callback_mode/0,terminate/3]).
 -export([get_server_name_by_imsi/1, get_pid_by_imsi/1]).
 -export([auth_request/2, lu_request/1, tunnel_request/2, purge_ms_request/1]).
@@ -48,12 +49,15 @@
          state_wait_create_session_resp/3, state_wait_delete_session_resp/3,
          state_wait_swm_session_termination_answer/3, state_active/3]).
 
+-define(TIMEOUT_VAL_WAIT_GTP_ANSWER, 10000).
+
 -record(ue_fsm_data, {
         imsi,
-        apn = "internet" :: string(),
-        tun_pdp_ctx :: epdg_tun_pdp_ctx,
-        tear_down_gsup_needed = false :: boolean(), %% need to send GSUP PurgeMSResp after STR+STA?
-        tear_down_gsup_cause = 0 :: integer()
+        apn                     = "internet"    :: string(),
+        pgw_rem_addr_list       = []            :: list(),
+        tun_pdp_ctx                             :: epdg_tun_pdp_ctx,
+        tear_down_gsup_needed   = false         :: boolean(), %% need to send GSUP PurgeMSResp after STR+STA?
+        tear_down_gsup_cause    = 0             :: integer()
         }).
 
 get_server_name_by_imsi(Imsi) ->
@@ -64,13 +68,18 @@ get_pid_by_imsi(Imsi) ->
         ServerName = get_server_name_by_imsi(Imsi),
         whereis(ServerName).
 
-start_link(Imsi) ->
+start_monitor(Imsi) ->
         ServerName = get_server_name_by_imsi(Imsi),
-        lager:info("ue_fsm start_link(~p)~n", [ServerName]),
-        gen_statem:start_link({local, ServerName}, ?MODULE, Imsi, [{debug, [trace]}]).
+        lager:info("ue_fsm start_monitor(~p)~n", [ServerName]),
+        gen_statem:start_monitor({local, ServerName}, ?MODULE, Imsi, [{debug, [trace]}]).
 
 stop(SrvRef) ->
-        gen_statem:stop(SrvRef).
+        try
+                gen_statem:stop(SrvRef)
+        catch
+        exit:Err ->
+                {error, Err}
+        end.
 
 auth_request(Pid, {PdpTypeNr, Apn}) ->
         lager:info("ue_fsm auth_request~n", []),
@@ -167,6 +176,12 @@ received_gtpc_delete_bearer_request(Pid) ->
 %% Internal helpers
 %% ------------------------------------------------------------------
 
+ev_handle({call, From}, {auth_request, PdpTypeNr, Apn}, Data) ->
+        case epdg_diameter_swm:auth_request(Data#ue_fsm_data.imsi, PdpTypeNr, Apn) of
+        ok -> {next_state, state_wait_auth_resp, Data, [{reply,From,ok}]};
+        {error, Err} -> {stop_and_reply, Err, Data, [{reply,From,{error,Err}}]}
+	end.
+
 %% ------------------------------------------------------------------
 %% gen_statem Function Definitions
 %% ------------------------------------------------------------------
@@ -190,12 +205,9 @@ terminate(Reason, State, Data) ->
 state_new(enter, _OldState, Data) ->
         {keep_state, Data};
 
-state_new({call, From}, {auth_request, PdpTypeNr, Apn}, Data) ->
+state_new({call, _From} = EvType, {auth_request, PdpTypeNr, Apn} = EvContent, Data) ->
         lager:info("ue_fsm state_new event=auth_request {~p, ~p}, ~p~n", [PdpTypeNr, Apn, Data]),
-        case epdg_diameter_swm:auth_request(Data#ue_fsm_data.imsi, PdpTypeNr, Apn) of
-        ok -> {next_state, state_wait_auth_resp, Data, [{reply,From,ok}]};
-        {error, Err} -> {stop_and_reply, Err, Data, [{reply,From,{error,Err}}]}
-	end;
+        ev_handle(EvType, EvContent, Data);
 
 state_new({call, From}, purge_ms_request, Data) ->
         lager:info("ue_fsm state_new event=purge_ms_request, ~p~n", [Data]),
@@ -204,10 +216,10 @@ state_new({call, From}, purge_ms_request, Data) ->
 state_wait_auth_resp(enter, _OldState, Data) ->
         {keep_state, Data};
 
-state_wait_auth_resp({call, From}, {received_swm_auth_response, Auth}, Data) ->
-        lager:info("ue_fsm state_wait_auth_resp event=received_swm_auth_response, ~p~n", [Data]),
-        gsup_server:auth_response(Data#ue_fsm_data.imsi, Auth),
-        case Auth of
+state_wait_auth_resp({call, From}, {received_swm_auth_response, Result}, Data) ->
+        lager:info("ue_fsm state_wait_auth_resp event=received_swm_auth_response Result=~p, ~p~n", [Result, Data]),
+        gsup_server:auth_response(Data#ue_fsm_data.imsi, Result),
+        case Result of
                 {ok, _} ->
                         {next_state, state_authenticating, Data, [{reply,From,ok}]};
                 {error, Err} ->
@@ -219,6 +231,10 @@ state_wait_auth_resp({call, From}, {received_swm_auth_response, Auth}, Data) ->
 state_authenticating(enter, _OldState, Data) ->
         {keep_state, Data};
 
+state_authenticating({call, _From} = EvType, {auth_request, PdpTypeNr, Apn} = EvContent, Data) ->
+        lager:info("ue_fsm state_authenticating event=auth_request {~p, ~p}, ~p~n", [PdpTypeNr, Apn, Data]),
+        ev_handle(EvType, EvContent, Data);
+
 state_authenticating({call, From}, lu_request, Data) ->
         lager:info("ue_fsm state_authenticating event=lu_request, ~p~n", [Data]),
         % Rx "GSUP CEAI LU Req" is our way of saying Rx "Swm Diameter-EAP REQ (DER) with EAP AVP containing successuful auth":
@@ -233,13 +249,22 @@ state_authenticating({call, From}, {received_swm_auth_compl_response, Result}, D
         lager:info("ue_fsm state_authenticating event=lu_request, ~p, ~p~n", [Result, Data]),
         % Rx "GSUP CEAI LU Req" is our way of saying Rx "Swm Diameter-EAP REQ (DER) with EAP AVP containing successuful auth":
         case Result of
-                {ok, _} ->
-                        Ret = ok;
-                {error, Err} ->
-                        Ret = {error, Err}
-        end,
-        gsup_server:lu_response(Data#ue_fsm_data.imsi, Ret),
-        {next_state, state_authenticated, Data, [{reply,From,Ret}]}.
+        {ok, ResInfo} ->
+                % Store PGW Remote address if AAA/HSS signalled them to us:
+                case maps:find(pdp_info_list, ResInfo) of
+                error ->
+                        Data1 = Data;
+                PGWAddrCandidateList ->
+                        Data1 = Data#ue_fsm_data{pgw_rem_addr_list = PGWAddrCandidateList}
+                end,
+                Ret = ok,
+                gsup_server:lu_response(Data1#ue_fsm_data.imsi, Ret),
+                {next_state, state_authenticated, Data1, [{reply,From,Ret}]};
+        {error, Err} ->
+                Ret = {error, Err},
+                gsup_server:lu_response(Data#ue_fsm_data.imsi, Ret),
+                {next_state, state_new, Data, [{reply,From,Ret}]}
+        end.
 
 state_authenticated(enter, _OldState, Data) ->
         {keep_state, Data};
@@ -250,7 +275,10 @@ state_authenticated({call, _From}, {auth_request, PdpTypeNr, Apn}, Data) ->
 
 state_authenticated({call, From}, {tunnel_request, PCO}, Data) ->
         lager:info("ue_fsm state_authenticated event=tunnel_request, ~p~n", [Data]),
-        epdg_gtpc_s2b:create_session_req(Data#ue_fsm_data.imsi, Data#ue_fsm_data.apn, PCO),
+        epdg_gtpc_s2b:create_session_req(Data#ue_fsm_data.imsi,
+                                         Data#ue_fsm_data.apn,
+                                         PCO,
+                                         Data#ue_fsm_data.pgw_rem_addr_list),
         {next_state, state_wait_create_session_resp, Data, [{reply,From,ok}]};
 
 state_authenticated({call, From}, purge_ms_request, Data) ->
@@ -264,16 +292,16 @@ state_authenticated({call, From}, received_gtpc_delete_bearer_request, Data) ->
         Data1 = Data#ue_fsm_data{tear_down_gsup_needed = false},
         {next_state, state_wait_swm_session_termination_answer, Data1, [{reply,From,ok}]};
 
-state_authenticated({call, From}, _Whatever, Data) ->
-        lager:error("ue_fsm state_authenticated: Unexpected call event, ~p~n", [Data]),
+state_authenticated({call, From}, Event, Data) ->
+        lager:error("ue_fsm state_authenticated: Unexpected call event ~p, ~p~n", [Event, Data]),
         {keep_state, Data, [{reply,From,ok}]};
 
-state_authenticated(cast, _Whatever, Data) ->
-        lager:error("ue_fsm state_authenticated: Unexpected cast event, ~p~n", [Data]),
+state_authenticated(cast, Event, Data) ->
+        lager:error("ue_fsm state_authenticated: Unexpected cast event ~p, ~p~n", [Event, Data]),
         {keep_state, Data}.
 
 state_wait_create_session_resp(enter, _OldState, Data) ->
-        {keep_state, Data, [{state_timeout,5000,create_session_timeout}]};
+        {keep_state, Data, [{state_timeout,?TIMEOUT_VAL_WAIT_GTP_ANSWER,create_session_timeout}]};
 
 state_wait_create_session_resp({call, From}, {received_gtpc_create_session_response, Result}, Data) ->
         lager:info("ue_fsm state_authenticated event=received_gtpc_create_session_response, ~p~n", [Data]),
@@ -288,11 +316,14 @@ state_wait_create_session_resp({call, From}, {received_gtpc_create_session_respo
                                               eua = EUA, peer_addr = RemoteIPv4},
                 Ret = gtp_u_tun:create_pdp_context(TunPdpCtx),
                 lager:debug("gtp_u_tun:create_pdp_context(~p) returned ~p~n", [ResInfo, Ret]),
-                Data1 = Data#ue_fsm_data{tun_pdp_ctx = TunPdpCtx};
-        _ -> Data1 = Data
-        end,
-        gsup_server:tunnel_response(Data1#ue_fsm_data.imsi, Result),
-        {next_state, state_active, Data1, [{reply,From,ok}]};
+                Data1 = Data#ue_fsm_data{tun_pdp_ctx = TunPdpCtx},
+                gsup_server:tunnel_response(Data1#ue_fsm_data.imsi, Result),
+                {next_state, state_active, Data1, [{reply,From,ok}]};
+        {error, GtpCause} ->
+                GsupCause = conv:cause_gtp2gsup(GtpCause),
+                gsup_server:tunnel_response(Data#ue_fsm_data.imsi, {error, GsupCause}),
+                {next_state, state_authenticated, Data, [{reply,From,ok}]}
+        end;
 
 state_wait_create_session_resp({call, From}, Event, Data) ->
         lager:error("ue_fsm state_wait_delete_session_resp: Unexpected call event ~p, ~p~n", [Event, Data]),
@@ -300,7 +331,7 @@ state_wait_create_session_resp({call, From}, Event, Data) ->
 
 state_wait_create_session_resp(state_timeout, create_session_timeout, Data) ->
         lager:error("ue_fsm state_wait_create_session_resp: Timeout ~p, ~p~n", [create_session_timeout, Data]),
-        gsup_server:tunnel_response(Data#ue_fsm_data.imsi, {error, create_session_timeout}),
+        gsup_server:tunnel_response(Data#ue_fsm_data.imsi, {error, ?GSUP_CAUSE_CONGESTION}),
         {next_state, state_authenticated, Data}.
 
 state_active(enter, _OldState, Data) ->
@@ -328,12 +359,12 @@ state_active({call, From}, received_gtpc_delete_bearer_request, Data) ->
         Data1 = Data#ue_fsm_data{tun_pdp_ctx = undefined, tear_down_gsup_needed = false},
         {next_state, state_wait_swm_session_termination_answer, Data1, [{reply,From,ok}]};
 
-state_active({call, From}, _Whatever, Data) ->
-        lager:error("ue_fsm state_active: Unexpected call event, ~p~n", [Data]),
+state_active({call, From}, Event, Data) ->
+        lager:error("ue_fsm state_active: Unexpected call event ~p, ~p~n", [Event, Data]),
         {keep_state, Data, [{reply,From,ok}]};
 
-state_active(cast, _Whatever, Data) ->
-        lager:error("ue_fsm state_active: Unexpected cast event, ~p~n", [Data]),
+state_active(cast, Event, Data) ->
+        lager:error("ue_fsm state_active: Unexpected cast event ~p, ~p~n", [Event, Data]),
         {keep_state, Data}.
 
 state_wait_delete_session_resp(enter, _OldState, Data) ->
